@@ -9,8 +9,22 @@
  * revised 31 January 1996. That document updates RFC1589.
  *
  * Copyright (c) 1996, 1997 by Ulrich Windl
- * $Id: enable_pps.c,v 3.6 1997/09/25 21:02:00 windl Exp $
- */
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ * $Id: enable_pps.c,v 1.4 1997/10/19 22:03:52 windl Exp $ */
 #include	<unistd.h>
 #include	<time.h>
 #include	<stdlib.h>
@@ -23,7 +37,7 @@
 #include	<sys/ioctl.h>
 #include	<termios.h>
 
-static	const char	id[] = "$Id: enable_pps.c,v 3.6 1997/09/25 21:02:00 windl Exp $";
+static	const char	id[] = "$Id: enable_pps.c,v 1.4 1997/10/19 22:03:52 windl Exp $";
 
 static	char	polarity	= 'p';	/* p[ositive] | n[egative] | 0) */
 
@@ -44,18 +58,54 @@ void	print_stats(const struct timex *tp)
 
 enum pps_state { S_NO_PPS, S_PPS_TIME, S_PPS_FREQ };
 
+static	char const	*status_str[] = {	/* from <timex.h> */
+	"PLL",		/* bit 0 */
+	"PPSFREQ",	/* bit 1 */
+	"PPSTIME",	/* bit 2 */
+	"FLL",		/* bit 3 */
+
+	"INS",		/* bit 4 */
+	"DEL",		/* bit 5 */
+	"UNSYNC",	/* bit 6 */
+	"FREQHOLD",	/* bit 7 */
+
+	"PPSSIGNAL",	/* bit 8 */
+	"PPSJITTER",	/* bit 9 */
+	"PPSWANDER",	/* bit 10 */
+	"PPSERROR",	/* bit 11 */
+
+	"CLOCKERR",	/* bit 12 */
+	"???13",	/* bit 13 */
+	"???14",	/* bit 14 */
+	"???15",	/* bit 15 */
+};
+
+/* write strings for set status bits on ``fp'' */
+static	void	write_status_bits(FILE *fp, int bits)
+{
+	int	b;
+	char	*sep = "";
+	for ( b = 0; b < 16; ++b )
+	{
+		if ( (bits & 1) != 0 )
+			fprintf(fp, "%s%s", sep, status_str[b]), sep = ", ";
+		bits >>= 1;
+	}
+}
+
 /* print PPS status */
-static	void	pps_status(int status)
+static	void	pps_status(int retval, int status)
 {
 	struct ppsclockev	event;
 	static	struct ppsclockev	last;
-	static	char	tag[] = "PPS: ";
-	char	*sep	= tag;
+	static	char const	tag[] = "PPS: ";
+	char const	*state_str = NULL;
+	char const	*sep	= tag;
 
 	if ( ioctl(fd, CIOGETEV, &event) == -1 )
 	{
 		perror("ioctl(CIOGETEV)");
-		return;
+		memset(&event, 0, sizeof(event));
 	}
 	
 	if ( event.serial != last.serial )
@@ -74,14 +124,25 @@ static	void	pps_status(int status)
 				sep, evs, secs), sep = ", ";
 		last = event;
 	}
-	if ( status & STA_PPSJITTER )
-		fprintf(stderr, "%sjitter exceeded", sep), sep = ", ";
-	if ( status & STA_PPSWANDER )
-		fprintf(stderr, "%swander exceeded", sep), sep = ", ";
-	if ( status & STA_PPSERROR )
-		fprintf(stderr, "%scalibration error", sep), sep = ", ";
+	switch ( retval )
+	{
+	case TIME_INS: state_str = "TIME_INS"; break;
+	case TIME_DEL: state_str = "TIME_DEL"; break;
+	case TIME_OOP: state_str = "TIME_OOP"; break;
+	case TIME_WAIT: state_str = "TIME_WAIT"; break;
+	case TIME_ERROR: state_str = "TIME_ERROR"; break;
+	}
+	if ( state_str != NULL )
+		fprintf(stderr, "%sstate=%s", sep, state_str), sep = ", ";
+
 	if ( !(status & STA_PPSSIGNAL) )
 		fprintf(stderr, "%sno pulse", sep), sep = ", ";
+	status &= (STA_PPSJITTER|STA_PPSERROR|STA_PPSWANDER);
+	if ( status != 0 )
+	{
+		fprintf(stderr, "%s", sep), sep = ", ";
+		write_status_bits(stderr, status);
+	}
 	if ( sep != tag )
 		fprintf(stderr, "\n");
 }
@@ -92,8 +153,7 @@ static	int	verbose_adjtimex(struct timex *txp)
 	int	state = adjtimex(txp);
 	if ( state > 0 )
 	{
-		fprintf(stderr, "time state = %d\n", state);
-		pps_status(txp->status);
+		pps_status(state, txp->status);
 	}
 	else if ( state < 0 )
 	{
@@ -107,13 +167,29 @@ static	int	verbose_adjtimex(struct timex *txp)
 static	int	update_status(struct timex *txp, int clear_bits, int set_bits)
 {
 	int	old_status = txp->status;
+	int	change;
+	char	*sep = " (";
 
 	txp->status |= set_bits;
 	txp->status &= ~clear_bits;
 	txp->modes = MOD_STATUS;
-	if ( txp->status != old_status )
-		fprintf(stderr, "update status: %04x -> %04x\n",
-			old_status, txp->status);
+	if ( txp->status == old_status )
+		return(0);
+	fprintf(stderr, "update status: %04x -> %04x",
+		old_status, txp->status);
+	if ( (change = ~old_status & (old_status ^ txp->status)) != 0 )
+	{
+		fprintf(stderr, "%sset=[", sep), sep = ", ";
+		write_status_bits(stderr, change);
+		fprintf(stderr, "]");
+	}
+	if ( (change = old_status & (old_status ^ txp->status)) != 0 )
+	{
+		fprintf(stderr, "%sclear=[", sep), sep = ", ";
+		write_status_bits(stderr, change);
+		fprintf(stderr, "]");
+	}
+	fprintf(stderr, "%s\n", sep[0] == ',' ? ")" : "");
 	return( verbose_adjtimex(txp) );
 }
 
@@ -148,6 +224,12 @@ static	void	pll_adj_offset(struct timex *txp)
 	verbose_adjtimex(txp);
 }
 
+/* print status message */
+static	void	status(char const *str)
+{
+	fprintf(stderr, "[%s]\n", str);
+}
+
 /* provide a usage message */
 static	void	usage(void)
 {
@@ -241,8 +323,8 @@ int	main(int argc, char *argv[])
 	state = S_NO_PPS;
 	while ( monitor )
 	{
-		time_state = verbose_adjtimex(&tx);
-		pps_status(tx.status);
+		if ( (time_state = verbose_adjtimex(&tx)) == 0 )
+			pps_status(time_state, tx.status);
 		print_stats(&tx);
 		switch ( state )
 		{
@@ -254,6 +336,7 @@ int	main(int argc, char *argv[])
 				  && tx.stabil < MAXFREQ )
 			{
   				state = S_PPS_FREQ;
+				status("valid PPS signal");
 			}
 			if ( (tx.status & STA_PPSSIGNAL) != 0 )
 			{
@@ -266,12 +349,12 @@ int	main(int argc, char *argv[])
 				      STA_PPSFREQ | STA_PLL);
 			if ( time_state != 0 )
 			{
-				pps_status(tx.status);
+				status("bad PPS signal");
 				state = S_NO_PPS;
 			}
 			else if ( tx.stabil < (10 << SHIFT_USEC) )
 			{
-				fprintf(stderr, "PPS frequency stabilized\n");
+				status("PPS frequency stabilized");
 				state = S_PPS_TIME;
 			}
 			break;
@@ -280,9 +363,11 @@ int	main(int argc, char *argv[])
 				      STA_PPSTIME | STA_PPSFREQ | STA_PLL);
 			if ( time_state != 0 )
 			{
-				pps_status(tx.status);
 				if ( tx.stabil > (50 << SHIFT_USEC) )
+				{
+					status("PPS lost stability");
 					state = S_PPS_FREQ;
+				}
 			}
 			break;
 		}
